@@ -121,6 +121,68 @@ where
         self.atoms.iter()
     }
 
+    fn path_to_node(&self, node_id: NodeId) -> Vec<Op> {
+        let mut path = Vec::new();
+        let mut current = node_id;
+        loop {
+            let node = &self.nodes[current];
+            if let Some(op) = &node.operator {
+                path.push(op.clone());
+            }
+            let Some(parent) = node.parent else {
+                break;
+            };
+            current = parent;
+        }
+        path.reverse();
+        path
+    }
+
+    /// return all atoms, with the path of operators from the root
+    ///
+    /// Example:
+    /// ```
+    /// let input = "!A & !(B | C)";
+    /// let mut expr = bet::BeTree::new();
+    /// for c in input.chars() {
+    ///     match c {
+    ///         '&' | '|' | '!' => expr.push_operator(c),
+    ///         ' ' => {}
+    ///         '(' => expr.open_par(),
+    ///         ')' => expr.close_par(),
+    ///         _ => expr.push_atom(c),
+    ///     }
+    /// }
+    /// assert_eq!(
+    ///     expr.paths_to_atoms(),
+    ///     vec![
+    ///         (vec!['&', '!'], 'A'),
+    ///         (vec!['&', '!', '|'], 'B'),
+    ///         (vec!['&', '!', '|'], 'C'),
+    ///     ]
+    /// );
+    ///
+    /// ```
+    pub fn paths_to_atoms(&self) -> Vec<(Vec<Op>, Atom)> {
+        let mut paths_to_atoms = Vec::new();
+        for (node_id, node) in self.nodes.iter().enumerate() {
+            if let Child::Atom(atom_id) = node.left {
+                let path = self.path_to_node(node_id);
+                paths_to_atoms.push((path, self.atoms[atom_id].clone()));
+            }
+            if let Child::Atom(atom_id) = node.right {
+                let path = self.path_to_node(node_id);
+                paths_to_atoms.push((path, self.atoms[atom_id].clone()));
+            }
+        }
+        paths_to_atoms
+    }
+
+    /// tell whether the given operator is used in the expression
+    pub fn has_operator(&self, operator: &Op) -> bool {
+        self.nodes.iter().any(|node| node.operator.as_ref() == Some(operator))
+    }
+
     /// returns a reference to the last atom if it's the last
     /// pushed token. Return none in other cases (including
     /// when no atom has been pushed at all)
@@ -178,10 +240,21 @@ where
     pub fn push_atom(&mut self, atom: Atom) {
         self.last_pushed = TokenType::Atom;
         let atom_idx = self.store_atom(atom);
-        if self.nodes[self.tail].left.is_some() {
-            self.nodes[self.tail].right = Child::Atom(atom_idx);
-        } else {
+        if self.nodes[self.tail].left.is_none() {
             self.nodes[self.tail].left = Child::Atom(atom_idx);
+            return;
+        }
+        let mut current = self.tail;
+        loop {
+            if self.nodes[current].right.is_none() {
+                self.nodes[current].right = Child::Atom(atom_idx);
+                return;
+            }
+            let Some(parent) = self.nodes[current].parent else {
+                // no parent for {current}, atom is useless
+                return;
+            };
+            current = parent;
         }
     }
 
@@ -247,6 +320,7 @@ where
         let Some(parent_idx) = self.nodes[new_idx].parent else {
             // the replaced node was the head
             self.head = new_idx;
+            self.tail = new_idx;
             return;
         };
         if self.nodes[parent_idx].left == Child::Node(self.tail) {
@@ -258,7 +332,6 @@ where
             self.nodes[parent_idx].right = Child::Node(new_idx);
         }
         // we connect the tail to the new node
-        //if let Child::Node(child_idx) = self.nodes[self.tail]I
         self.nodes[self.tail].parent = Some(new_idx);
         // and we update the tail
         self.tail = new_idx;
@@ -294,8 +367,7 @@ where
     /// operator at this point (for example it makes no
     /// sense just after another operator)
     pub fn accept_binary_operator(&self) -> bool {
-        use TokenType::*;
-        matches!(self.last_pushed, Atom | ClosingPar)
+        matches!(self.last_pushed, TokenType::Atom | TokenType::ClosingPar)
     }
 
     /// tell whether it would make sense to push an atom
@@ -438,7 +510,7 @@ where
         };
         if short_circuit(op, &left_value) {
             return Ok(Some(left_value));
-        };
+        }
         let right_value =
             self.eval_child_faillible(eval_atom, eval_op, short_circuit, node.right)?;
         Ok(Some(eval_op(op, left_value, right_value)?))
